#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pytoxr contains tools for the analysis of data from ToxR experiments

Copyright (C) 2016  Mark George Teese

This software is licensed under the permissive MIT License.
"""

# figures plotted for each experimen1t can be put here.


from pytoxr.parse import parse_softmax
from pytoxr.parse import parse_all_data_files_in_folder
from pytoxr.collect import collect_data_in_folder
import pytoxr.toolkit
import pytoxr.scanmut
import pytoxr.fit_to_sine
from pytoxr.scanmut.scanmut import run_scanmut
from pytoxr.parse import extract_mut_names_from_collected_data
import pytoxr.mathfunctions
import pytoxr.collect
import pytoxr.parse


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pytoxr contains tools for the analysis of data from ToxR experiments

Copyright (C) 2016  Mark George Teese

This software is licensed under the permissive MIT License.
"""
import glob
import matplotlib.pyplot as plt
import os
import pandas as pd
import sys
import numpy as np
from scipy.stats import ttest_ind

def collect_data_in_folder(target_dir):
    """ Collect all ToxR data in a directory, and save in a subdirectory "collected".

    Parameters
    ----------
    target_dir : str
        Directory with .pda, .txt and .xls input files
        Also should contain subdirectories with parsed data from SoftMax Pro to excel/csv tables.

    """
    sys.stdout.write("Starting collect_data_in_folder.\n")
    sys.stdout.flush()

    collect_out_dir = os.path.join(target_dir, "collected")
    if not os.path.isdir(collect_out_dir):
        os.makedirs(collect_out_dir)
    collected_excel = os.path.join(collect_out_dir, "collected.xls")

    # target_dir = r"B:\xbackups\MarkTeese\Praktikum\2017 MEM & MEM PROTEINS\2017_data\old_toxr_results_sep_folders\group3"
    txt_file_list = glob.glob(os.path.join(target_dir, "*.txt"))


    n = 0
    analysed_txt_file_list = []
    exp_name_list = []
    df = pd.DataFrame()
    dfnu_list = []

    for txt_file in txt_file_list:
        print(txt_file)
        filename = os.path.basename(txt_file)
        if "excluded" in filename:
            # skip this file, do not include with other datapoints
            print("{} skipped (has 'excluded' in the filename)".format(filename))
            continue
        pda_file_path = txt_file[:-4] + ".pda"
        samples_file_path = txt_file[:-4] + ".xls"
        exp_name = os.path.basename(txt_file)[:-4][0:60]
        out_dir = os.path.join(os.path.dirname(txt_file), exp_name)
        # for compatibility purposes, currently only excel 2003 is recommended
        excel_format = ".xls"
        out_parsed_excel = os.path.join(out_dir, "{}_parsed{}".format(exp_name, excel_format))
        # check if pda and samples files exist
        if os.path.isfile(pda_file_path) and os.path.isfile(samples_file_path):
            pass
        else:
            # skip this text file
            continue

        # check if parsed excel file exists
        if not os.path.isfile(out_parsed_excel):
            print("{} skipped, no analysed data found".format(exp_name))
            continue

        # open dataframe normalised unique
        dfnu = pd.read_excel(out_parsed_excel, sheetname="norm_unique")
        # rename columns to have a prefix that shows the file name
        newcols = ["{}||{}".format(n, c) for c in dfnu.columns]
        dfnu.columns = newcols
        # add dataframe, filename and experiment names to lists
        dfnu_list.append(dfnu)
        analysed_txt_file_list.append(txt_file)
        exp_name_list.append(exp_name)

        n += 1

    # join all dfnu dataframes
    df = pd.concat(dfnu_list, axis=1)

    """
    Dataframe df should now look like this.

    Input files are labelled 0 to 8, for example. Columns are renamed to input_file_number||col_heading

                     0||order  0||error    0||OD1    0||OD2    0||OD3  ...    \
    P02724-0_GpA_wt    1        NaN        0.893932  0.856831  1.030202  ...
    AZ2                2        NaN        0.747473  0.753499  0.569780  ...
    P02724-0_GpA_G83A  3        NaN        0.878750  0.798697  1.018703  ...
    dTM                4        NaN        0.772738  1.039832  1.467613  ...
    Q6ZRP7-2_QSOX2_wt  5        NaN        1.000000  1.000000  1.000000  ...

                        8||MU2    8||MU3  8||MU_mean  8||MU_std  8||Vm
    P02724-0_GpA_wt    0.718398  0.820212  0.743628    0.417787   Vm01
    AZ2                1.070686 -1.716484  0.440827    16.463324  Vm17
    P02724-0_GpA_G83A  0.375756  0.337701  0.359586    0.538124   Vm25
    dTM               -0.029046  0.027206 -0.017234    0.347562   Vm02
    Q6ZRP7-2_QSOX2_wt  1.000000  1.000000  1.000000    1.000000   NaN
    """

    # copy across the sample order from the original samples file. since some norm-unique data is missing samples,
    # the orig sample order is most likely the max of the available sample orders
    order_cols = ["{}||order".format(n) for n in range(n)]
    df_order = df[order_cols].copy()
    order_ser = df_order.max(axis=1)
    df["order"] = order_ser
    # sort by the sample order, for the following mean dataframes and figures
    df.sort_values("order", inplace=True)

    ttest_standard_cols = ["{}||ttest_standard".format(n) for n in range(n)]
    #ttest_standard_columns = set(ttest_standard_cols).intersection(set(df.index))
    ttest_standard_columns = set(ttest_standard_cols).intersection(set(df.columns))

    # WEIRD NON-REPEATABLE DF.T EFFECT TROUBLESHOOTING
    # print("ttest_standard_cols", ttest_standard_cols)
    # print("ttest_standard_columns", ttest_standard_columns)
    # print("df.index", df.index)
    # print("df.columns", df.columns)

    # check that there are actually some ttest standards labelled
    if ttest_standard_columns != set():
        conduct_ttest = True
        df_ttest_standard = df[ttest_standard_cols].copy()
        df_ttest_standard = df_ttest_standard.dropna(how="all")
        if df_ttest_standard.shape[0] == 1:
            ttest_standard = df_ttest_standard.index[0]
        else:
            raise ValueError("seems to be more than one ttest_standard selected in the original sample excel files."
                             "df_ttest_standard.index = {}".format(df_ttest_standard.index))
    else:
        conduct_ttest = False

    data_type_list = ["{}||OD_mean", "{}||Vi_mean", "{}||MU_mean"]
    writer = pd.ExcelWriter(collected_excel)

    data_is_normalised = True
    dfs = pd.read_excel(samples_file_path)
    standard_subset = dfs.loc[dfs.standard == True]
    standard_sample_name = standard_subset["sample"].unique()[0]
    if data_is_normalised:
        norm_string = ", normalised to {}".format(standard_sample_name)
    else:
        norm_string = ""
    titles_dict = {"OD" : "OD600, all experiments", "Vi" : "Initial Velocity (Vi), all experiments", "MU" : "Miller Units, all experiments"}
    y_axis_label_dict = {"OD": "OD600{}".format(norm_string), "Vi": "Initial Velocity (Vi, U/min){}".format(norm_string), "MU": "Miller Units (U/min/OD600){}".format(norm_string)}


    for data_type in data_type_list:

        excel_tab_name = data_type.split("||")[1].split("_")[0]
        title = titles_dict[excel_tab_name]
        y_axis_label = y_axis_label_dict[excel_tab_name]

        scatter_path = os.path.join(collect_out_dir, "{}_scatter_plot.png".format(excel_tab_name))
        box_path = os.path.join(collect_out_dir, "{}_box_plot.png".format(excel_tab_name))

        selected_cols = [data_type.format(n) for n in range(n)]
        df_sel = df[selected_cols].copy()
        df_sel.columns = exp_name_list
        df_sel.to_excel(writer, sheet_name=excel_tab_name)

        """
        df_sel looks like this, and only shows the normalised miller units

        2010915_prak_samples  group1_day1  group1_day2  \
        P02724-0_GpA_wt    0.814460              0.951131     1.001340
        AZ2                2.377500              2.795199     5.627452
        P02724-0_GpA_G83A  0.233870              0.262983     0.389499
        dTM                0.473288              0.048931     0.027387
        Q6ZRP7-2_QSOX2_wt  1.000000              1.000000     1.000000
        """

        plt.close("all")
        ax = df_sel.plot(marker="o", linestyle="", alpha=0.5, figsize=(8, 12))
        ax.set_xticks(range(df_sel.shape[0]))
        ax.set_xticklabels(df_sel.index, rotation=90)
        #ax.set_ylim(0, 6)
        ax.set_xlim(-1, df_sel.shape[0] + 1)
        ax.set_title(title)
        ax.set_ylabel(y_axis_label)
        plt.legend()
        plt.tight_layout()
        plt.savefig(scatter_path, dpi=240)
        plt.savefig(scatter_path[:-4] + ".pdf")
        plt.close("all")

        # pandas describe function doesn't seem to have an axis, so this clumsy method is used to flip columns and index twice
        df_sel_describe = df_sel.T.describe().T
        #df_sel_describe["std"] = df_sel_describe["std"].fillna(0)
        df_sel_describe["sem"] = df_sel_describe["std"] / np.sqrt(df_sel_describe["count"])
        df_sel_describe["order"] = order_ser

        print("conduct_ttest = {}".format(conduct_ttest))

        sample_list = df_sel.index.tolist()

        for sample_name in sample_list:
            data = df_sel.loc[sample_name, :].dropna()
            # get the median. Note that the median of [1,2,3,40] is 2.5.
            df_sel_describe.loc[sample_name, "median"] = data.median()

            if conduct_ttest and sample_name != ttest_standard:
                # get the standard "wildtype" data for the TTEST
                ttest_standard_data = df_sel.loc[ttest_standard, :]
                #sample_list.remove(ttest_standard)

                if len(data) > 1 and len(ttest_standard_data) > 1:
                    p_value = ttest_ind(ttest_standard_data, data)[1]
                else:
                    p_value = np.nan
                df_sel_describe.loc[sample_name, "ttest_p_value"] = p_value

            df_sel_describe["sign_stars"] = df_sel_describe["ttest_p_value"].apply(return_p_value_stars)
            df_sel_describe["ttest_significant"] = df_sel_describe["ttest_p_value"] < 0.05
        df_sel_describe.to_excel(writer, sheet_name=excel_tab_name + "_describe")

        max_value = df_sel_describe["max"].max()
        min_value = df_sel_describe["min"].max()

        """
        df_describe now looks like this

                           count      mean       std       min       25%       50%       75%       max       sem  order    median  ttest_p_value sign_stars  ttest_significant
        P02724-0_GpA_wt     14.0  1.000000  0.000000  1.000000  1.000000  1.000000  1.000000  1.000000  0.000000    1.0  1.000000       0.000073        ***               True
        AZ2                 13.0  1.163821  0.314529  0.964444  1.037625  1.065475  1.173914  2.167819  0.087235    2.0  1.065475       0.918356        NaN              False
        P02724-0_GpA_G83A   12.0  1.246863  0.466768  0.840600  1.052239  1.132847  1.236926  2.672585  0.134744    3.0  1.132847       0.578728        NaN              False
        dTM                 13.0  1.372324  0.462053  1.015609  1.067898  1.100609  1.624869  2.466696  0.128151    4.0  1.100609       0.135864        NaN              False
        Q6ZRP7-2_QSOX2_wt   14.0  1.173372  0.137769  0.926538  1.134087  1.155864  1.232545  1.403153  0.036820    5.0  1.155864            NaN        NaN              False

        """
        # plot barchart with two averaging methods, mean and median
        for stat_method in ["mean", "median"]:

            bar_path = os.path.join(collect_out_dir, "{}_{}_bar_plot.png".format(excel_tab_name, stat_method))
            plt.close("all")
            # fig = plt.Figure(figsize=(5,10))
            ax = df_sel_describe[stat_method].plot(kind="bar", figsize=(8, 12), yerr=df_sel_describe['sem'])
            #
            # sign_ = df_sel_describe["sign_stars"].copy()
            # sign_ser.index = range(sign_ser.shape[0])
            # sign_ser.dropna(inplace=True)


            if conduct_ttest:
                df_sel_describe["#"] = range(1, df_sel_describe.shape[0] + 1)
                sign_df = df_sel_describe.dropna(subset=["sign_stars"])

                for sample in sign_df.index:
                    x = sign_df.loc[sample, "#"] - 1
                    y = sign_df.loc[sample, stat_method]
                    sign_star = sign_df.loc[sample, "sign_stars"]
                    ax.annotate(sign_star, [x - 0.18, y + 0.005], horizontalalignment='center')
            ax.set_title(title)
            ax.set_ylabel(y_axis_label)
            #ax.errorbar(range(df_sel_describe.shape[0]), df_sel_describe['mean'], yerr=df_sel_describe['sem'])

            plt.tight_layout()
            plt.savefig(bar_path, dpi=240)
            plt.savefig(bar_path[:-4] + ".pdf")
            plt.close("all")

            meanpointprops = dict(marker='o', markerfacecolor='black', markersize=2)  # markeredgecolor='0.75',
            fig, ax = plt.subplots(figsize=(7, 10))
            df_sel.T.plot(kind="box", whis=1500, showmeans=True, meanprops=meanpointprops, showfliers=True, ax=ax)
            ax.set_xticklabels(df_sel.index, rotation=90)
            ax.set_title(title)
            ax.set_ylabel(y_axis_label)

            if conduct_ttest:
                for sample in sign_df.index:
                    x = sign_df.loc[sample, "#"]
                    y = sign_df.loc[sample, "75%"]
                    sign_star = sign_df.loc[sample, "sign_stars"]
                    ax.annotate(sign_star, [x - 0.18, y + 0.005], horizontalalignment='center')
            if data_is_normalised:
                if max_value > 6:
                    ylim_min = ax.get_ylim()[0]
                    ax.set_ylim(ylim_min,6)
            plt.tight_layout()
            plt.savefig(box_path, dpi=240)
            plt.savefig(box_path[:-4] + ".pdf")
            plt.close("all")

    sys.stdout.write("\n\ncollect_data_in_folder is finished.\Data from {} files collected in total.\n-----------------------------------------------------------------\n".format(n+1))
    sys.stdout.flush()

    writer.close()

def return_p_value_stars(p):
    if float(p) < 0.001:
        return "***"
    elif float(p) < 0.01:
        return "**"
    elif float(p) < 0.05:
        return "*"
    else:
        return np.nan


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pytoxr contains tools for the analysis of data from ToxR experiments

Copyright (C) 2016  Mark George Teese

This software is licensed under the permissive MIT License...
"""
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import csv

def create_heatmap_from_softmax_endpoint(datafile, skiprows, skipfooter, startcol, endcol, vmin):
    """

    Parameters
    ----------
    datafile
    skiprows
    skipfooter
    startcol
    endcol

    Returns
    -------

    Usage
    -----
    datafile = r"D:\data\toxr_pipette_test_Brigitte_Thermomax.txt"
    create_heatmap_from_softmax_endpoint(datafile, skiprows=6, skipfooter=103, startcol=2, endcol=14, vmin=0.075)

    """
    plt.style.use('seaborn-whitegrid')
    heatmap_path = datafile[:-4] + "_heatmap.png"
    df = pd.read_csv(datafile, skiprows=skiprows, skipfooter=skipfooter, sep="\t", engine="python")
    df = df.iloc[:, startcol:endcol]
    plt.close("all")
    fig, ax = plt.subplots()
    ax = sns.heatmap(df, ax=ax, vmin=vmin)
    ax.get_figure()
    # rotate the y-ticklabels
    ax.set_yticklabels(ax.get_yticklabels(), rotation=0)
    # set the x-axis labels on the top
    ax.xaxis.tick_top()
    fig.savefig(heatmap_path, dpi=240)

def aaa(df_or_series):
    """ Function for use in debugging.
    Saves pandas Series or Dataframes to a user-defined csv file.
    """
     # convert any series to dataframe
    if isinstance(df_or_series, pd.Series):
        df_or_series = df_or_series.to_frame()
    csv_out = r"D:\data\000_aaa_temp_df_out.csv"
    df_or_series.to_csv(csv_out, sep=",", quoting=csv.QUOTE_NONNUMERIC)


def extract_aa_pos_from_sample_names(df_with_sn_as_index, start_aa, end_aa, newcol_name):
    """Extract the amino acid position from sample names in the index of a dataframe.

    Note this cannot be used if the data contains double-mutants. Consider using a dictionary approach instead.

    Parameters
    ----------
    df_with_sn_as_index : dataframe with the sample name as the index
        Dataframe to be altered, will be returned with the new column
    start_aa : int, starting amino acid in range of amino acids mutated
        UniProt style, NOT python indexing style
    end_aa : int, end amino acid in range of amino acids mutated
        UniProt style where the number is the last aa to be included, NOT python indexing style
    newcol_name : name of the new column containing the index

    Returns
    -------
    df_with_sn_as_index : original dataframe, with the new column
    """

    for i in range(start_aa,end_aa+1):
        # for each sample name in the dataframe with unique samples
        for sn in df_with_sn_as_index.index:
            # if the amino acid number (233, 234, etc) is in the sample name
            if str(i) in sn:
                # add the amino acid number to a new column
                df_with_sn_as_index.loc[sn, newcol_name] = i



def calc_av_over_3aa_window(df, data_col, new_col, n_aa = 4):
    ''' Function to average the properties of an amino acid (conservation, disruption, etc) across a window, consisting
    of three amino acids on the same side of an alpha helix (i, i-4, i+4)
    INPUTS:
    df: dataframe
    data_col: string showing the column name, which contains the data to be averaged
    new_col: string for the new column name, containing the average
    n_aa: number of amino acids separating the aa. Default = 4 (e.g. i, i-4, i+4)
    OUTPUTS:
    The original dataframe, with an extra column.
    '''
    import numpy as np
    # iterate over each row of the dataframe
    for aa_pos in df.index:
        # define the central datapoint
        disrupt_centre = df.loc[aa_pos,data_col]
        # define the "left" N-term datapoint (e.g. i-4)
        if aa_pos-n_aa in df.index:
            disrupt_left = df.loc[aa_pos-n_aa,data_col]
        else:
            disrupt_left = np.nan
        # define the "right" C-term datapoint (e.g. i+4)
        if aa_pos+n_aa in df.index:
            disrupt_right = df.loc[aa_pos+n_aa,data_col]
        else:
            disrupt_right = np.nan
        # calc average
        av_disrupt = np.nanmean([disrupt_centre, disrupt_left, disrupt_right])
        #print(aa_pos, disrupt_centre, disrupt_left, disrupt_right)
        # add average to dataframe
        df.loc[aa_pos,new_col] = av_disrupt
    return df


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pytoxr contains tools for the analysis of data from ToxR experiments

Copyright (C) 2016  Mark George Teese

This software is licensed under the permissive MIT License...
"""
import glob
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import re
import seaborn as sns
import sys

def parse_softmax(txt_path, samples_path):
    """Parse the text file output from Sofmax Pro, using the standard A600 endpoint & A405 kinetic template.
    This method uses the initial velocities calculated by Softmax Pro.

    Parameters
    ----------
    txt_path : str
        Path to txt data file.
    samples_path : str
        Path to an excel file with all of the sample names.
        An example of this text file should be in the "examples" subfolder.

    Dataframes
    ----------
    df = raw softmax pro output
    dfd = dataframe for data.
        index = Vm#, columns = Vi1      Vi2     Vi3       MU1       MU2, etc
    dfnu = dataframe for normalised, unique data
        index = sample_name, columns = as above for dfd

    Returns
    -------
    dfd : pd.DataFrame
        Dataframe for data.
    """
    exp_name = os.path.basename(txt_path)[:-4][0:60]
    out_dir = os.path.join(os.path.dirname(txt_path), exp_name)
    # path to output heatmap of raw OD600 values
    out_OD600_heatmap = os.path.join(out_dir, "{}_OD600_heatmap.png".format(exp_name))
    # path to excel output file
    # for compatibility purposes, currently only excel 2003 is recommended
    excel_format = ".xls"
    out_parsed_excel = os.path.join(out_dir, "{}_parsed{}".format(exp_name, excel_format))
    # only the non-normalised data is saved as csv
    out_parsed_csv = os.path.join(out_dir, "{}_parsed.csv".format(exp_name))

    # create output directory
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)


    """
    FORMAT OF THE OD600 section of the SOFTMAX PRO FILE

~End
Plate:	OD600	1,3	PlateFormat	Endpoint	Absorbance	Raw	TRUE	1						1	600	1	12	96	1	8	None
	Temperature(¡C)	1	2	3	4	5	6	7	8	9	10	11	12
	24,10	0,1094	0,1322	0,121	0,117	0,1214	0,1239	0,1128	0,1219	0,1191	0,1152	0,1172	0,1164
		0,153	0,1564	0,1582	0,1518	0,1636	0,1528	0,1448	0,1651	0,1371	0,1484	0,1491	0,1509
		0,1194	0,1218	0,1266	0,12	0,1171	0,1252	0,1155	0,1227	0,123	0,1204	0,1221	0,1159
		0,1217	0,1237	0,1239	0,119	0,1217	0,1245	0,1168	0,1241	0,1207	0,1168	0,1203	0,1203
		0,1152	0,119	0,1402	0,1184	0,1443	0,1219	0,1193	0,1254	0,1206	0,1173	0,1167	0,1165
		0,1253	0,1313	0,1435	0,1232	0,1261	0,1298	0,1239	0,1315	0,1133	0,1193	0,1157	0,1178
		0,1136	0,1143	0,1359	0,1172	0,1373	0,1275	0,1159	0,1281	0,1224	0,1195	0,1168	0,1143
		0,1078	0,1206	0,1243	0,1139	0,1199	0,1229	0,1172	0,121	0,1206	0,0379	0,0382	0,0407
    """

    # go through each line, searching for the various components
    OD600_regex_search_string = "Plate:\sOD600"
    # regex_search_string = "Group:\s+Results Kinetic\s+1"
    OD600_plate_is_found = False
    OD600_plate_line = 0
    first_line_with_OD600_data_int = None
    first_line_with_OD600_data = None



    """
    CHECK IF "OD600" or "Cell Density (OD600)" is used in this template.
    Plate:	Cell Density (OD600)	1,3	PlateFormat	Endpoint	Absorbance	Raw	TRUE	1						1	600	1	12	96	1	8	None
        Temperature(¡C)	1	2	3	4	5	6	7	8	9	10	11	12
        28,00	0,109467	0,119067	0,119167	0,112067	0,140267	0,126567	0,043267	0,052867	0,053967	0,052567	0,052567	0,052767

    """
    OD600_regex_search_string = None
    with open(txt_path, "r") as f:
        for line in f:
            if "Cell Density (OD600)" in line:
                OD600_regex_search_string = "Plate:\sCell\sDensity\s\(OD600\)"
                break
            if "Plate:\tOD600" in line:
                OD600_regex_search_string = "Plate:\sOD600"
                break
        if OD600_regex_search_string is None:
            raise ValueError("Line with OD600 not found in text template. Check text output file.")

    with open(txt_path, "r") as f:
        for n, line in enumerate(f):
            match = re.match(OD600_regex_search_string, line)
            if match:
                OD600_plate_is_found = True
                OD600_plate_line = n
                first_line_with_OD600_data_int = OD600_plate_line + 2
            if n == first_line_with_OD600_data_int:
                first_line_with_OD600_data = line
    total_n_lines = n

    if not OD600_plate_is_found:
        raise ValueError("Line with OD600 not found in text template. Check text output file.")

    table_start = first_line_with_OD600_data_int
    table_end = first_line_with_OD600_data_int + 8
    skipfooter = total_n_lines - table_end + 1

    if "," in first_line_with_OD600_data:
        decimal = ","
    elif "." in first_line_with_OD600_data:
        decimal = "."
    df = pd.read_table(txt_path, skiprows=table_start, sep='\s+', decimal=decimal, header=None, skipfooter=skipfooter, engine="python")

    """
    OD600 data now looks something like this.
    The temperature in row 0 puts everything out of order.

            0       1       2       3       4       5       6       7       8        9       10      11      12
    0  24.1000  0.1094  0.1322  0.1210  0.1170  0.1214  0.1239  0.1128  0.1219   0.1191  0.1152  0.1172  0.116
    1   0.1530  0.1564  0.1582  0.1518  0.1636  0.1528  0.1448  0.1651  0.1371   0.1484  0.1491  0.1509     Na
    2   0.1194  0.1218  0.1266  0.1200  0.1171  0.1252  0.1155  0.1227  0.1230   0.1204  0.1221  0.1159     Na
    3   0.1217  0.1237  0.1239  0.1190  0.1217  0.1245  0.1168  0.1241  0.1207   0.1168  0.1203  0.1203     Na
    4   0.1152  0.1190  0.1402  0.1184  0.1443  0.1219  0.1193  0.1254  0.1206   0.1173  0.1167  0.1165     Na
    5   0.1253  0.1313  0.1435  0.1232  0.1261  0.1298  0.1239  0.1315  0.1133   0.1193  0.1157  0.1178     Na
    6   0.1136  0.1143  0.1359  0.1172  0.1373  0.1275  0.1159  0.1281  0.1224
    7   0.1078  0.1206  0.1243  0.1139  0.1199  0.1229  0.1172  0.1210  0.1206   0.1195  0.1168  0.1143     Na
    """
    # fix up alignment in dataframe, rename columns and index
    df.loc[0, :] = list(df.loc[0, 1:]) + [0]
    df = df.loc[:, 0:11]
    df.index = list("ABCDEFGH")
    df.columns = range(1, 13)

    """
    OD600 data is ready for heatmap

               1       2       3       4       5       6       7       8       9   \
    A  0.1094  0.1322  0.1210  0.1170  0.1214  0.1239  0.1128  0.1219  0.1191
    B  0.1530  0.1564  0.1582  0.1518  0.1636  0.1528  0.1448  0.1651  0.1371
    C  0.1194  0.1218  0.1266  0.1200  0.1171  0.1252  0.1155  0.1227  0.1230
    D  0.1217  0.1237  0.1239  0.1190  0.1217  0.1245  0.1168  0.1241  0.1207
    E  0.1152  0.1190  0.1402  0.1184  0.1443  0.1219  0.1193  0.1254  0.1206
    F  0.1253  0.1313  0.1435  0.1232  0.1261  0.1298  0.1239  0.1315  0.1133
    G  0.1136  0.1143  0.1359  0.1172  0.1373  0.1275  0.1159  0.1281  0.1224
    H  0.1078  0.1206  0.1243  0.1139  0.1199  0.1229  0.1172  0.1210  0.1206

    """
    tum_blue4_as_python_color = np.array([0, 82, 147]) / 255
    cmap = sns.light_palette(tum_blue4_as_python_color, as_cmap=True)

    plt.close("all")
    fig, ax = plt.subplots()
    ax = sns.heatmap(df, ax=ax, cmap=cmap, vmin=0.075)
    ax.get_figure()
    # rotate the y-ticklabels
    ax.set_yticklabels(ax.get_yticklabels(), rotation=0)
    # set the x-axis labels on the top
    ax.xaxis.tick_top()
    fig.savefig(out_OD600_heatmap, dpi=240 )

    """
    OD600 heatmap is finished.
    Now grab the kinetic data, which is presented in a table at the bottom of the text file.

    This is the FORMAT OF THE INPUT SOFTMAX PRO FILE, at region of processed kinetic data

    ..... <----------- raw kinetic data is above. we currently use the sofmax pro calculated initial velocity
    ~End
    Group:	Results Kinetic	1                                <----------- group_results_line. Use regex to find the line number associated with this string.
    Sample	Wells	Sample#	V/max	Vmax/OD600	Mean	SD
    Vm01	A1	1	43,605	1231,785	1379,232	132,621  <----------- first_line_with_data. Use this to determine if . or , is the decimal.
        A2		59,253	1488,774
        A3		50,733	1417,135
    Vm02	B1	2	26,652	870,965	801,480	75,053
        B2		39,200	811,590
        B3		36,816	721,885

    ...

    ~End The last line with ~End shows the location of the end of the file. This is variable. Best to use dropna() to drop empty rows at the end of the dataframe.
    """

    # go through each line, searching for the various components
    regex_search_string = "Group:\s+Results Kinetic\s+1"
    group_results_are_found = False
    group_results_line = 0
    first_line_with_data_int = None
    first_line_with_data = None
    start_of_table = None

    with open(txt_path, "r") as f:
        for n, line in enumerate(f):
            match = re.match(regex_search_string, line)
            if match:
                group_results_are_found = True
                group_results_line = n
                first_line_with_data_int = group_results_line + 2
                start_of_table = group_results_line+1
            if n == start_of_table:
                start_of_table_data = line
            if n == first_line_with_data_int:
                first_line_with_data = line

    if not group_results_are_found:
        raise ValueError("Softmax template is not recognised. Check input file names, and that 'Group: Results Kinetic' is located in the text file.")

    # check if the data is saved using a German or English decimal point
    if "," in first_line_with_data:
        dec = ","
    elif "." in first_line_with_data:
        dec = "."
    else:
        raise TypeError("VersaMax exported text file seems to have an unknown decimal format. Try re-exporting data.")

    skiprows = group_results_line

    # DEPRECATED. DID NOT WORK.
    # if "Sample" in start_of_table_data:
    #     skiprows = group_results_line
    # elif "Group" in start_of_table_data:
    #     skiprows = group_results_line + 1
    # else:
    #     raise ValueError("skiprows does not work. check txt file format.")

    # open csv as a pandas dataframe, starting from "Sample	Wells	Sample#	V/max	Vmax/OD600	Mean" ...etc
    df = pd.read_csv(txt_path, sep='\t', skiprows=skiprows, decimal=dec)

    if 'Sample' not in df.columns:
        """
        .
        pandas versions seem to use a different skiprows method
        need to check that the skiprows is correctly aligned
        Group:	Results Kinetic	1
        Sample	Wells	Sample#	V/max	Vmax/OD600	Mean	SD
        Vm01	A1	1	55,264	790,996	776,806	48,225
        """
        if "Group:" in df.columns:
            skiprows = group_results_line + 1
            df = pd.read_csv(txt_path, sep='\t', skiprows=skiprows, decimal=dec)

        else:
            raise ValueError("Pandas has tried to make a dataframe from OD600 values, but can't seem to locate the header.\n"
                             "Check text file for output.\nCurrent dataframe columns : {}.".format(df.columns))

    # drop the last footer rows, which may vary in length
    df.dropna(subset=["V/max"], inplace=True)
    # fill the names downwards, so that they can be used for creating pivot tables later
    df.fillna(method="pad", inplace=True)

    # search for datapoints that could not be fitted
    if "NoFit" in df["V/max"].tolist():
        df_no_fit = df.loc[df["V/max"] == "NoFit"].copy()
        df_no_fit["Vm_Well"] = df_no_fit["Sample"] + "_well_" + df_no_fit["Wells"]
        sample_names_with_no_fit = df_no_fit["Sample"].unique()
        samples_with_no_fit = df_no_fit["Vm_Well"].dropna().unique()
        sys.stdout.write("'NoFit' found in data, indicating measurement was impossible, probably due to very high enzyme concentrations.\n"
                         "Samples affected: {}".format(samples_with_no_fit))

        df.replace("NoFit", np.nan, inplace=True)
        df["V/max"] = df["V/max"].astype(float)
        df["Vmax/OD600"] = df["Vmax/OD600"].astype(float)

        #df["V/max"] = df["V/max"].replace("NoFit", np.nan).astype(float)
        #df["Vmax/OD600"] = df["Vmax/OD600"].replace("NoFit", np.nan).astype(float)
        # drow any rows with no fit
        #df = df.loc[df["V/max"].notnull()]
    else:
        sample_names_with_no_fit = []

    # calculate OD from V and V/OD (not currently supplied by the template)
    df["OD"] = df["V/max"] / df["Vmax/OD600"]


    """
    df should look like this. Basically, it looks like the SoftMax pro output using Jan Kirrbach's template.

      Sample Wells  Sample#   V/max  Vmax/OD600      Mean       SD        OD
    0   Vm01    A1      1.0  43.605    1231.785  1379.232  132.621  0.035400
    1    NaN    A2      NaN  59.253    1488.774       NaN      NaN  0.039800
    2    NaN    A3      NaN  50.733    1417.135       NaN      NaN  0.035800
    3   Vm02    B1      2.0  26.652     870.965   801.480   75.053  0.030601
    4    NaN    B2      NaN  39.200     811.590       NaN      NaN  0.048300
    """

    Vi_replicates = ["Vi1","Vi2","Vi3"] * int(df.shape[0]/3)
    MU_replicates = ["MU1","MU2","MU3"] * int(df.shape[0]/3)
    OD_replicates = ["OD1","OD2","OD3"] * int(df.shape[0]/3)
    df["Vi_rep"] = Vi_replicates
    df["MU_rep"] = MU_replicates
    df["OD_rep"] = OD_replicates

    dfv = df.pivot_table(values="V/max", index = "Sample", columns="Vi_rep")
    dfd = df.pivot_table(values="Vmax/OD600", index = "Sample", columns="MU_rep")
    dfo = df.pivot_table(values="OD", index = "Sample", columns="OD_rep")
    dfd = pd.concat([dfv, dfd, dfo], axis=1)

    """
    dfd now contains the values for each replicate, one after another

               Vi1      Vi2     Vi3       MU1       MU2       MU3       OD1     OD2     OD3
    Sample
    Vm01    43.605   59.253  50.733  1231.785  1488.774  1417.135  0.035400  0.0398  0.0358
    Vm02    26.652   39.200  36.816   870.965   811.590   721.885  0.030601  0.0483  0.0510
    Vm03    36.813   66.878  56.552  1191.375  1706.078  1516.140  0.030900  0.0392  0.0373
    Vm04    42.418   46.726  45.454  1258.694  1115.172  1014.602  0.033700  0.0419  0.0448
    Vm05    70.692  101.895  88.205  2380.192  3078.387  2183.284  0.029700  0.0331  0.0404

    Vi is the initial velocity.
    MU is the miller units
    OD is the OD600 (cell density)

    now the mean and standard deviation is simply recalculated for the Vi, MU and OD replicates
    """

    dfd["Vi_mean"] = dfd.loc[:, "Vi1":"Vi3"].mean(axis=1)
    dfd["Vi_std"] = dfd.loc[:, "Vi1":"Vi3"].std(axis=1)
    dfd["MU_mean"] = dfd.loc[:, "MU1":"MU3"].mean(axis=1)
    dfd["MU_std"] = dfd.loc[:, "MU1":"MU3"].std(axis=1)
    dfd["OD_mean"] = dfd.loc[:, "OD1":"OD3"].mean(axis=1)
    dfd["OD_std"] = dfd.loc[:, "OD1":"OD3"].std(axis=1)


    # open excel file with sample names
    dfs = pd.read_excel(samples_path, index_col=0)

    dfs["frame"] = dfs.frame.dropna().astype(int).astype(str)
    dfs.fillna("", inplace=True)
    for i in dfs.index:
        dfs.loc[i, "sample_python"] = dfs.loc[i, "uniprot":"mut"].dropna().astype(str).str.cat()
        # try:
        #     dfs.loc[i, "sample_python"] = dfs.loc[i, "uniprot":"mut"].dropna().astype(str).str.cat()
        # except:
        #     print(i)
        #     print(dfs.loc[i, "uniprot":"mut"])
        #     print(dfs.loc[i, "uniprot":"mut"].dropna())
        #     dfs.loc[i, "sample_python"] = dfs.loc[i, "uniprot":"mut"].dropna().astype(str).str.cat()
    if dfs["sample"].tolist() != dfs.sample_python.tolist():
        raise ValueError("sample column in excel does not exactly match data in preceding columns. Try repeating fill-down.")
    sample_name_dict = dfs.set_index("Vm#")["sample"].to_dict()
    order_dict = dfs.set_index("Vm#")["order"].to_dict()
    include_data_dict = dfs.set_index("Vm#")["include_data"].to_dict()

    """
    dfs should now look like this
        Vm# uniprot  - frame  _ shortname __  mut              sample      notes    include_data       sample_python
    #                                                                             #
    1  Vm01  P02724  -     0  _       GpA  _   wt     P02724-0_GpA_wt        GpA  1         True     P02724-0_GpA_wt
    2  Vm02     ΔTM                                               ΔTM        ΔTM  2         True                 ΔTM
    3  Vm03  Q6ZRP7  -     2  _     QSOX2  _  L4A  Q6ZRP7-2_QSOX2_L4A  QSOX2_L4A  3         True  Q6ZRP7-2_QSOX2_L4A
    4  Vm04  Q6ZRP7  -     2  _     QSOX2  _  S8A  Q6ZRP7-2_QSOX2_S8A  QSOX2_S8A  4         True  Q6ZRP7-2_QSOX2_S8A
    5  Vm05  Q6ZRP7  -     2  _     QSOX2  _  S8Q  Q6ZRP7-2_QSOX2_S8Q  QSOX2_S8Q  5         True  Q6ZRP7-2_QSOX2_S8Q
    """

    dfd["Vm"] = dfd.index
    dfd["sample"] = dfd["Vm"].replace(sample_name_dict)
    dfd["order"] = dfd["Vm"].replace(order_dict)
    dfd["include_data"] = dfd["Vm"].replace(include_data_dict)

    # note that there has been an error in some samples
    dfd.loc[sample_names_with_no_fit, "error"] = "NoFit"
    dfd["error"] = dfd["error"].fillna("none")

    # change the column order
    col_order = ['sample', 'order','include_data', 'error', 'OD1', 'OD2', 'OD3','OD_mean', 'OD_std','Vi1', 'Vi2', 'Vi3', 'Vi_mean', 'Vi_std', 'MU1', 'MU2', 'MU3', 'MU_mean', 'MU_std', 'Vm']
    dfd = dfd.reindex(columns=col_order)
    dfd.sort_values("order", inplace=True)

    dfn = pd.DataFrame()

    if True in dfs.standard.tolist():
        standard_subset = dfs.loc[dfs.standard == True]
        standard_sample_name_list = standard_subset.sample_python.unique()
        if len(standard_sample_name_list) != 1:
            raise ValueError("Multiple standards are selected. Double-check excel file with sample names.")
        standard_sample_name = standard_sample_name_list[0]
        standard_Vm_numbers = standard_subset.index.tolist()

        standard_values_ser_or_df = dfd.loc[dfd["sample"] == standard_sample_name]
        if isinstance(standard_values_ser_or_df, pd.DataFrame):
            standard_mean_values_ser = standard_values_ser_or_df.mean(axis=0)
        elif isinstance(standard_values_ser_or_df, pd.Series):
            standard_mean_values_ser = standard_values_ser_or_df
        else:
            raise TypeError("Oops! 'standard_values_ser_or_df' is neither a dataframe nor a series")

        for i in range(dfd.shape[0]):
            row_series = dfd.iloc[i, :]
            Vm = dfd.index[i]
            dfn[Vm] = row_series / standard_mean_values_ser

        dfn = dfn.T

        """ Norm dataframe has problem in that the Std is 1 for the standard. (norm to standard, not std/mean_of_standard)
        This is fixed as below.
                    OD2        OD3   OD_mean     OD_std        Vi1        Vi2  \
        Vm01          1          1         1          1          1          1   
        Vm17    1.09716   0.692331  0.966009    3.70093    3.08974    2.61241   
        Vm25    0.19164     1.3053  0.566228    8.62627   0.360652  -0.164633   
        Vm02    1.30523    1.44421   1.41189  0.0739331   0.045963  0.0785936   
        """
        datatypes = ["OD", "Vi", "MU"]
        for datatype in datatypes:
            std_column = dfd["{}_std".format(datatype)]
            std_normed = std_column / standard_mean_values_ser["{}_mean".format(datatype)]
            # replace the column in dfn
            dfn["{}_std".format(datatype)] = std_normed

        """std is corrected, and should now look like this (a value other than 1 for the standard!)
        
                    Vi3    Vi_mean    Vi_std   Vm error include_data order sample  
        Vm01          1          1  0.117211  NaN   NaN            1     1    NaN  
        Vm17    2.56545    2.75133  0.369209  NaN   NaN            1     2    NaN  
        Vm25   0.201911   0.114587  0.273156  NaN   NaN            0     3    NaN  
        """

        for transferred_col in ["sample", "order", "Vm", "include_data"]:
            dfn[transferred_col] = dfd[transferred_col]
        dfn = dfn.reindex(columns=col_order)
        dfn.sort_values("order", inplace=True)

        """
        dfn still has the Vm numbers as the index, which allows duplicates (e.g. two wells of QSOX2_wt in this experiment)

                          sample  order       OD1       OD2       OD3   OD_mean
        Vm01     P02724-0_GpA_wt      1  0.801497  0.853056  0.767382    0.8074
        Vm17                 AZ2      2  0.880439  0.935936  0.531282  0.779956
        Vm25   P02724-0_GpA_G83A      3  0.185782   0.16348   1.00166  0.457173
        Vm02                 dTM      4   1.20262   1.11343   1.10826   1.13996
        Vm09   Q6ZRP7-2_QSOX2_wt      5   1.05084  0.940833  0.935106  0.973782
        Vm24   Q6ZRP7-2_QSOX2_wt      6  0.949156   1.05917   1.06489   1.02622
        Vm10  Q6ZRP7-2_QSOX2_C1S      7   1.42985   1.56866     1.648   1.55175
        Vm18  Q6ZRP7-2_QSOX2_V2A      8   1.40206   1.50719   1.54405   1.48644
        Vm26  Q6ZRP7-2_QSOX2_V3A      9   1.11713  0.116117   1.39081  0.869624
        Vm03  Q6ZRP7-2_QSOX2_L4A     10  0.931424  0.765291  0.812374  0.834106
        """

        unique_samples = dfn["sample"].unique()
        dfn.set_index("sample", inplace=True)

        # exclude datapoints according to manual annotation in the "include_data" column
        dfn_include_data = dfn.loc[dfn.include_data]

        if dfn_include_data.empty:
            # all data has been labelled for exclusion here
            sys.stdout.write()
            sys.stdout.write("\n'{}' skipped, all data labeled as FALSE for 'include_data' in samples excel file.".format(exp_name))
            # skip this protein
            return dfd

        dfnu = pd.DataFrame()

        for s in unique_samples:
            if s not in dfn_include_data.index:
                # skip this unique sample, as it was manually excluded via the "include_data" column
                continue
            row_df_or_ser = dfn_include_data.loc[s, :]
            if isinstance(row_df_or_ser, pd.Series):
                dfnu.loc[:, s] = row_df_or_ser
            elif isinstance(row_df_or_ser, pd.DataFrame):
                # take the order of the first duplicate row as the correct one
                order = row_df_or_ser.iloc[0, :]["order"]
                # drop the text rows, so the dtype is changed to float
                # otherwise the mean of two values for the standard  (which should give 1.0) returns np.nan
                # reason for strange behaviour in pandas is unknown
                row_df_or_ser = row_df_or_ser.drop(["Vm", "order"], axis=1)
                mean_of_duplicate_samples = row_df_or_ser.mean(axis=0)
                # add the single row to the unique dataframe
                dfnu.loc[:, s] = mean_of_duplicate_samples
                dfnu.loc["order", s] = order

        dfnu = dfnu.T

        dropped_cols = ["include_data", "error"]
        dfnu.drop(dropped_cols, axis=1, inplace=True)

        ttest_standard_true_selection = dfs[dfs.ttest_standard == True]
        ttest_standard_samplename = ttest_standard_true_selection["sample"].unique()[0]
        if ttest_standard_samplename == standard_sample_name:
            sys.stdout.write("\n--------------------------------------------------------------------"
                             "Sorry, script is currently not designed to handle situations where"
                             "the TTEST standard is the same as the norm standard. TTEST results should be ignored. "
                             "Talk to Mark.------------------------------------------------------------------------\n")

        ttest_standard_array = dfs.loc[dfs.ttest_standard == True].sample_python.unique()
        if len(ttest_standard_array) == 1:
            ttest_standard = ttest_standard_array[0]
            if ttest_standard in dfnu.index:
                dfnu.loc[ttest_standard, "ttest_standard"] = True
        elif len(ttest_standard_array) > 1:
            raise ValueError("seems to be two different ttest standards selected")
        elif len(ttest_standard_array) == 0:
            print("no ttest_standard selected")

    """
    dfnu is now completely normalised to your standard (e.g. QSOX2).
    This includes all OD measurements, initial velocity, and Miller Units.
    The index is now unique. The values of duplicates show the mean for that experiment.
    The values of duplicate standards (e.g. Q6ZRP7-2_QSOX2_wt) are now 1.0.

                          order       OD1       OD2       OD3   OD_mean     OD_      Vi1        Vi2        Vi3    Vi_mean      Vi_std
    P02724-0_GpA_wt       1  0.928313  0.898545  0.556443     0.748  0.0632076  0.594462   0.683314   0.632995   0.637554    0.669004
    AZ2                   2  0.709755  0.673359  0.401063  0.556855          0  0.563122   0.530232   0.466218   0.516645           0
    P02724-0_GpA_G83A     3   5.40776   5.13045   3.05578   4.24278          0  0.778933   0.733438   0.644892   0.714644           0
    dTM                   4   3.29763   2.30533  0.788788   1.86296    2.09499 0.0715175  0.0668064  0.0597716  0.0656486  0.00467747
    Q6ZRP7-2_QSOX2_wt     5         1         1         1         1          1         1          1          1          1           1
    """

    dfd.index.name = "Vm"
    dfd.to_csv(out_parsed_csv)

    with pd.ExcelWriter(out_parsed_excel) as writer:
        dfd.to_excel(writer, sheet_name="original")
        dfn.to_excel(writer, sheet_name="normalised")
        dfnu.to_excel(writer, sheet_name="norm_unique")
        writer.close()

    sys.stdout.write("\n'{}' parsed successfully".format(exp_name))
    if skiprows == group_results_line + 1:
        sys.stdout.write("   (Note: You may have an older version of Pandas or python. A skiprows error has been averted.)")
    sys.stdout.flush()

    return dfd

def parse_all_data_files_in_folder(target_dir, reparse_existing=False):
    """Parse all ToxR SoftMax Pro txt files in a given directory.

    Parameters
    ----------
    target_dir : str
        Directory with .pda, .txt and .xls input files

    """
    txt_file_list = glob.glob(os.path.join(target_dir, "*.txt"))

    counter = 0

    for txt_file in txt_file_list:
        pda_file_path = txt_file[:-4] + ".pda"
        samples_file_path = txt_file[:-4] + ".xls"
        exp_name = os.path.basename(txt_file)[:-4][0:60]
        out_dir = os.path.join(os.path.dirname(txt_file), exp_name)
        excel_format = ".xls"
        out_parsed_excel = os.path.join(out_dir, "{}_parsed{}".format(exp_name, excel_format))


        if os.path.isfile(pda_file_path) and os.path.isfile(samples_file_path):
            if not reparse_existing:
                if os.path.isfile(out_parsed_excel):
                    # skip analysis of this file
                    sys.stdout.write("\n{} skipped, parsed output file already exists.".format(exp_name))
                    sys.stdout.flush()
                    continue

            dfd = parse_softmax(txt_file, samples_file_path)
            counter += 1

    sys.stdout.write("\n\nparse_all_data_files_in_folder is finished.\n{} file(s) parsed in total.\n-----------------------------------------------------------------\n".format(counter))
    sys.stdout.flush()


def contains_kw(string):
    control_keywords = ["GpA", "dTM", "AZ2"]
    for kw in control_keywords:
        if kw in string:
            return True
    return False

def extract_mut_names_from_collected_data(collected_xls, scan_mut_xlsx, stat_method="mean"):
    """Parses the collected.xls file with ToxR data so that it can be used as input
    for scanning mutagenesis barchart creation.

    Parameters
    ----------
    collected_xls : str
        Path to input collected.xls
    scan_mut_xlsx : str
        Path to output scan_mut.xlsx
    """

    df_coll = pd.read_excel(collected_xls, sheetname="Vi_describe")

    """ df_coll looks like this
                        count      mean       std       min       25%       50%       75%       max       sem  order
    P02724-0_GpA_wt         7  1.000000  0.000000  1.000000  1.000000  1.000000  1.000000  1.000000  0.000000      1
    AZ2                     7  0.761144  0.061944  0.631301  0.760551  0.775837  0.786084  0.827599  0.023413      2
    P02724-0_GpA_G83A       7  0.376627  0.204756  0.088955  0.252525  0.295542  0.566675  0.613493  0.077390      3
    dTM                     7  0.036521  0.021639  0.005306  0.028094  0.036204  0.040399  0.077153  0.008179      4
    Q6ZRP7-2_QSOX2_wt       7  0.906791  0.410194  0.528483  0.606863  0.811692  1.051753  1.690128  0.155039      5
    Q6ZRP7-2_QSOX2_C1S      5  0.880809  0.143449  0.773950  0.791562  0.836428  0.874484  1.127623  0.064153      7
    
    """
    # identify common controls and remove them from the sample list
    df_coll["name"] = df_coll.index
    df_coll["contains_kw"] = df_coll["name"].apply(contains_kw).tolist()
    df_coll = df_coll.loc[~df_coll["contains_kw"]].copy()

    # extract orig_aa, final_aa, aa_position etc from the full sample name (e.g. Q6ZRP7-2_QSOX2_C1S)
    df_coll["mut"] = df_coll["name"].str.split("_").str[-1]
    df_coll["orig_aa"] = df_coll["mut"].str[0]
    df_coll["target_aa"] = df_coll["mut"].str[-1]
    df_coll["aa_position_unit"] = df_coll["mut"].str[1:-1]

    # assume that any name that contains "wt" is the wildtype.
    wt_name = df_coll.loc[df_coll["name"].str.contains("wt")].index
    # IMPORTANT. Get the value for the wildtype. Used to normalise all values to wt later.
    wt_value = float(df_coll.loc[wt_name, stat_method])
    # if more than one name has wt, raise an error
    assert len(wt_name) == 1

    # rename some values in the the wildtype row
    df_coll.loc[wt_name, "orig_aa"] = "WT"
    df_coll.loc[wt_name, "target_aa"] = "B"
    df_coll.loc[wt_name, "aa_position_unit"] = "0"

    """df_coll now looks something like this:
                        count      mean       std       min       25%       50%       75%       max       sem  order                name  contains_kw  mut orig_aa target_aa aa_position_unit
    Q6ZRP7-2_QSOX2_wt       7  0.906791  0.410194  0.528483  0.606863  0.811692  1.051753  1.690128  0.155039      5   Q6ZRP7-2_QSOX2_wt        False   wt      WT         B                0
    Q6ZRP7-2_QSOX2_C1S      5  0.880809  0.143449  0.773950  0.791562  0.836428  0.874484  1.127623  0.064153      7  Q6ZRP7-2_QSOX2_C1S        False  C1S       C         S                1
    Q6ZRP7-2_QSOX2_V2A      5  0.737907  0.259606  0.553112  0.555503  0.569161  0.883270  1.128490  0.116099      8  Q6ZRP7-2_QSOX2_V2A        False  V2A       V         A                2
    """
    # take only the desired columns
    df = df_coll[["count", "mean", "median", "std", "sem", "orig_aa", "target_aa", "aa_position_unit", "ttest_p_value"]].copy()

    # normalise the mean, median, standard-deviation and SEM to the wildtype value
    for col in ["mean", "median", "std", "sem"]:
        df[col] = df[col] / wt_value

    # rename columns according to a dictionary
    rename_dict = {"std": "SD", "sem": "SE", "count": "Rep", stat_method: "perc_WT", "aa_position_unit": "amino acid position",
                   "orig_aa": "original amino acid", "target_aa": "final amino acid", "ttest_p_value" : "p_value"}
    df.rename(columns=rename_dict, inplace=True)
    # some graphs need a p-value
    if "p_value" not in df.columns:
        df["p_value"] = 1
    df.index = range(df_coll.shape[0])

    df.to_excel(scan_mut_xlsx)

    print("extract_mut_names_from_collected_data is finished")


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pytoxr contains tools for the analysis of data from ToxR experiments

Copyright (C) 2016  Mark George Teese

This software is licensed under the permissive MIT License.
"""

from setuptools import setup, find_packages
from os import path
from codecs import open

# grab the long_description from the readme file
here = path.abspath(path.dirname(__file__))
with open(path.join(here, "readme.rst")) as f:
    long_description = f.read()


classifiers = """\
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Programming Language :: Python :: 3
Topic :: Scientific/Engineering :: Bio-Informatics
Topic :: Scientific/Engineering :: Chemistry
"""

setup(name='pytoxr',
	author="Mark Teese",
    url="https://github.com/teese/pytoxr",
    download_url = 'https://github.com/teese/pytoxr/archive/0.0.6.tar.gz',
    author_email="mark.teese@checkmytumhomepage.de",
    description = "Tools for the analysis of data from ToxR experiments.",
    long_description=long_description,
	long_description_content_type='text/x-rst',
    license='MIT',
    packages=find_packages(),
    classifiers=classifiers.splitlines(),
    keywords="ToxR transmembrane TOXCAT TMDhomodimer GALLEX AraTM BacTH",
	project_urls={'LangoschLab':'http://cbp.wzw.tum.de/index.php?id=9', "TU_Munich":"https://www.tum.de"},
    install_requires=["pandas", "numpy", "matplotlib", "scipy", "seaborn", "eccpy"],
    version='0.0.6')


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author:         Mark Teese
Created:        06.11.2015
Dependencies:   Python 3.x
                Numpy
                SciPy
                Pandas
Purpose:        Fit a sine wave to experimental data
Credits:        All sections by Mark Teese.
"""
from __future__ import unicode_literals
import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize
import os
import pandas as pd
from korbinian.utils import create_colour_lists
import pytoxr.mathfunctions as mathfunctions

def fit_sin_to_data(x, y, title, output_folder, output_base_filename, fit_unbiased=True, fit_alphahelix=True,x_grid=True,savefig=True, savecsv=True, saveexcel=True):
    """ Fits a sinusoidal curve to data.

    1) fits orig data to unbiased sine curve
    2) fits orig data to sine curve biased for alpha helical periodicity of 3.6

    The output files are decided by the user, and can include a png figure, excel and/or csv files.

    Parameters
    ----------
    x : arraylike (e.g. numpy 1D array, pandas Series)
        Data of the x axis (e.g. aa position)
    y : arraylike (e.g. numpy 1D array, pandas Series)
        Array-like data of the y axis (e.g. disruption)
    title : string
        Title of the output figure
    output_folder : string
        Output path. E.g. r"D:\Data\fit_to_sine_output\"
    output_base_filename : string
        Start of filename. E.g. "20160607_experiment1"
    fit_unbiased : bool
        If True, the sinusoidal curve will be fit to the data with an unbiased periodicity.
    fit_alphahelix : bool
        If True, the sinusoidal curve will befit to the data with alpha-helical periodicity of 3.6 (constant b = 1.7)
    savefig : bool
        If True, output figures will be saved
    savecsv : bool
        If True, output csv files will be saved
    saveexcel : bool
        If True, output excel files will be saved

    Saved Files and Figures
    -----------------------
    basename.xlsx : excel file with the following tabs
        sine_constants : sine constants from fitted curves
        yvalues_fitted : x and y datapoints used to draw the curves
    basename_sine_constants.csv : csv
        sine constants from fitted curves
    basename_yvalues_fitted.csv : csv
        x and y datapoints used to draw the curves
    basename.png : line graph of original input data,
        fitted unbiased sine curve, and fitted sine curve with fixed periodicity

    Usage Example
    -------------
    import tlabtools as tools
    import pandas as pd
    '''
    LaPointe 2013 ToxR Ala scanning data of FtsB that was fitted to a sine curve in publication
    '''
    lit_data_xls = r"D:\PyCharmProjects\tlabtools\tlabtools\fit_to_sine\examples\P0A6S5.xlsx"
    output_folder = r"D:\DATABASES_spitfire\20160607_fit_to_sine_output"
    output_base_filename = "20160607_P0A6S5"
    # read data file
    dflit = pd.read_excel(lit_data_xls, sheetname="Sheet1")
    # define x
    ftsb_x = dflit.aa_position
    # define y (normalised between 0 and 1)
    ftsb_y = tools.normalise_0_1(dflit["Disruption(1-mutation)"])[0]
    # set title
    P0A6S5_title = "P0A6S5, FtsB (Senes Lab)"
    # run fit sin to data
    tools.fit_sin_to_data(ftsb_x,ftsb_y,P0A6S5_title,output_folder,output_base_filename,savefig=True)
    # REPEAT WITH RANDOMIZED/SHUFFLED DATA
    ftsb_y_shuffled = ftsb_y.copy()
    np.random.shuffle(ftsb_y_shuffled)
    output_base_filename_random = "20160607_P0A6S5_RANDOM"
    P0A6S5_title_random = "P0A6S5, FtsB (Senes Lab), RANDOM"
    # run fit sin to data
    tools.fit_sin_to_data(ftsb_x,ftsb_y_shuffled,P0A6S5_title_random,output_folder,output_base_filename_random,savefig=True)
    """
    if fit_unbiased == False and fit_alphahelix == False:
        raise ValueError("Either fit_unbiased or fit_alphahelix must be True")

    #global orig_data_df, df_sine_constants
    output_folder_pdf = os.path.join(output_folder, "pdf")
    output_basename = os.path.join(output_folder, output_base_filename)
    output_basename_pdf = os.path.join(output_folder_pdf, output_base_filename)

    # create output folders, if necessary
    list_paths = [output_folder, output_folder_pdf]
    for path in list_paths:
        if not os.path.exists(path):
            os.makedirs(path)

    # create a number of x data points for smooth sine curves
    x_smooth = np.linspace(x.min(), x.max(), 500)
    # create dataframe to save the sine constants
    df_sine_constants = pd.DataFrame(index = ["a","b","c","d","periodicity"])
    # create dataframe to save the fitted curves
    df_fitted_curve = pd.DataFrame(index=x_smooth)

    if fit_unbiased:
        #####################################################################################################
        #                                                                                                   #
        #              CALCULATE SINE FIT TO FREE HELIX (ALTHOUGH GUESS HAS PERIODICITY OF 3.6)             #
        #                                                                                                   #
        #####################################################################################################

        # guess the constants in the initial curve
        sine_constants_guess = [1.0,1.7,0.2,0]
        # fit a sine curve to the data using the leastsq method
        sine_constants, cov, infodict, mesg, ier = scipy.optimize.leastsq(mathfunctions.residuals,
                                                                          sine_constants_guess,
                                                                          args=(mathfunctions.sine,x,y),
                                                                          full_output=1)

        # print the periodicity of the curve that best fits the data (hopefully ~3.6, corresponding to an alpha-helix!)
        periodicity_fitted_curve = 2 * np.pi / sine_constants[1]
        print("periodicity of fitted curve = %0.2f" % periodicity_fitted_curve)
        # add the sine constants and periodicity to the output dataframe
        df_sine_constants["sine_constants_unbiased"] = sine_constants.tolist() + [periodicity_fitted_curve]
        # plot the fitted curve to the data
        yvalues_fitted = mathfunctions.sine(sine_constants, x_smooth)
        df_fitted_curve['yvalues_fitted_unbiased'] = yvalues_fitted

    if fit_alphahelix:
        #####################################################################################################
        #                                                                                                   #
        #                CALCULATE SINE FIT TO PERFECT ALPHA HELIX (FIXED PERIODICITY OF 3.6)               #
        #                                                                                                   #
        #####################################################################################################
        #RECALCULATE SINE CONSTANTS ASSUMING A PERFECT HELIX (a=0.2, b=1.7, periodicity = 3.6)
        sine_constants_guess_perfhelix = [1.0,0.5]
        # fit a sine curve to the data using the leastsq method
        sine_constants_A, cov_A, infodict_A, mesg_A, ier_A = scipy.optimize.leastsq(mathfunctions.residuals,
                                                                             sine_constants_guess_perfhelix,
                                                                             args=(mathfunctions.sine_perfect_helix,x,y),
                                                                             full_output=1)
        # add the sine constants to output dataframe
        df_sine_constants.loc["c":"d","sine_constants_alpha_helix"] = sine_constants_A
        # plot the fitted curve to the data
        yvalues_fitted_perfhelix = mathfunctions.sine_perfect_helix(sine_constants_A, x_smooth)
        df_fitted_curve['yvalues_fitted_fixed_periodicity'] = yvalues_fitted_perfhelix

    if x_grid:
        raise ValueError("ax not created?")
        #ax.grid(axis='x', which='major', alpha=0.5)

    if savefig:
        #####################################################################################################
        #                                                                                                   #
        #                                           SAVE OUTPUT FIGURES                                     #
        #                                                                                                   #
        #####################################################################################################
        # set some figure settings
        plt.rcParams["savefig.dpi"] = 120
        plt.rc('font', family='Arial')
        fontsize = 12
        # create custom list of colours for figures
        colour_lists = create_colour_lists()
        # create new fig
        fig, ax = plt.subplots()
        # plot the raw disruption data against the amino acid number
        ax.plot(x, y, color = colour_lists['TUM_colours']['TUMBlue'], label = "original data")
        if fit_unbiased:
            ax.plot(x_smooth, yvalues_fitted, linestyle = "--", color = colour_lists['TUM_colours']['TUM4'], label = "fit to sine (unbiased)")
            ax.annotate(s='periodicity = %0.1f' % periodicity_fitted_curve,
                        xy = (0.04,0.9),color = colour_lists['TUM_colours']['TUM4'],
                        fontsize=fontsize, xytext=None, xycoords='axes fraction',
                        alpha=0.75)

        if fit_alphahelix:
            ax.plot(x_smooth, yvalues_fitted_perfhelix, linestyle = "--", color = colour_lists['TUM_accents']['green'], label = "fit to sine (fixed periodicity)")
            ax.annotate(s='periodicity = 3.6',
                        xy = (0.04,0.8),color = colour_lists['TUM_accents']['green'],
                        fontsize=fontsize, xytext=None, xycoords='axes fraction',
                        alpha=0.75)

        raise ValueError("code needs to be rearranged if you want this to work. orig_aa not found.")
        set_ticks_and_labels_and_save(fig, ax, x, y, orig_aa, output_basename, output_basename_pdf, fontsize, title)

    if saveexcel:
        # save data to new excel file
        writer = pd.ExcelWriter(output_basename + ".xlsx")
        df_sine_constants.to_excel(writer, sheet_name = "sine_constants")
        df_fitted_curve.to_excel(writer, sheet_name='yvalues_fitted')
        writer.save()
        writer.close()
    if savecsv:
        df_sine_constants.to_csv(output_basename + "_sine_constants.csv")
        df_fitted_curve.to_csv(output_basename + "_yvalues_fitted.csv")

    # return sine_constants, periodicity_fitted_curve

def set_ticks_and_labels_and_save(fig, ax, x, y, orig_aa, output_basename, output_basename_pdf, fontsize, title):
    """ Set the ticks and other parameters in the output figure, and save as png and pdf

    Parameters
    ----------
    fig : figure object
    ax : ax (plot) object
    x : array-like data of the x axis (e.g. aa position)
    y : array-like data of the y axis (e.g. disruption)
    output_basename : base filepath of output figures
    output_basename_pdf : base filepath of output pdf figures
    fontsize : figure fontsize
    title : figure title

    Saved Files and Figures
    -------
    output_basename + ".png" : output figure, png
    output_basename_pdf + ".pdf" : output figure, pdf
    """

    """ Adjusts some of the figure parameters, and saves as a png and pdf using the basepath indicated
    """
    ax.set_xticks(x)
    ax.set_xticklabels(x, fontsize  = fontsize, rotation = 90)
    #use original aa as x axis label
    ax.set_xticklabels(orig_aa, fontsize  = fontsize, rotation = 0)
    ax.set_xlabel('orignal aa', fontsize=fontsize)
    ax.set_ylabel('average value', fontsize=fontsize)
    y_dist = max(y)-min(y)
    ylim_min = min(y) - y_dist*0.2 if min(y) - y_dist*0.2 > 0 else 0
    ylim_max = max(y) + y_dist*0.2
    #ax.set_ylim(min(y),max(y) + 0.4)
    ax.set_ylim(ylim_min, ylim_max + 0.4)
    ax.set_title(title, fontsize=fontsize)
    # show the legend(as previously labelled)
    ax.legend(fontsize=fontsize)
    #autoscale_x_axis
    plt.autoscale(enable=True, axis='x', tight=True)
    # save figure
    fig.savefig(output_basename + ".png")
    fig.savefig(output_basename_pdf + ".pdf", format="pdf")



def residuals(constants, function, x, y):
    """
    This is a value function (Kostenfunktion) that is used to optimise the fit of the curve to the data.
    It simply calculates the distance between y-value from real data and y-value from the function (sigmoid/sine/etc).
    """
    return y - function(constants, x)

def sine(sine_constants, x):
    ''' Sine equation used to fit sine curve to data.
    f (x) = a*sin (bx + c) + d
    see http://www.graphpad.com/guides/prism/6/curve-fitting/index.htm?reg_classic_dr_variable.htm
    '''
    import numpy as np
    a, b, c, d = sine_constants
    y = a * np.sin (b * x + c) + d
    return y

def sine_perfect_helix(sine_constants_cd, x):
    ''' Sine equation which is forced to retain alpha helical periodicity, with 3.6 residues per turn.
    f (x) = a*sin (bx + c) + d
    a = 0.2
    b = 1.745
    Why is a fixed to 0.2?
        This is arbitrary, resulting in a curve that is 0.4 in height
    Why is b fixed to 1.745?
        Since periodicity = 2*np.pi/a, for a perfect alpha helix b = 2*np.pi/3.6 = 1.745
    '''
    import numpy as np
    a = 0.2
    b = 1.745
    c, d = sine_constants_cd
    y = a * np.sin (b * x + c) + d
    return y


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import sys
import os
import ast
import eccpy.tools as eccpytools
import eccpy as eccpy
from pytoxr.toolkit import extract_aa_pos_from_sample_names
from pytoxr.fit_to_sine.fit_to_sine import fit_sin_to_data

def run_scanmut(settings_file, show_barchart=False, save_pdf=True):
    """ Runs the scanmut algorithm on the selected input files in the excel settings file.

    Scanmut takes typical scanning mutagenesis data, creates a barchart showing the values for each amino acid,
    and if desired, fits the data to a sine curve to judge if the data follows alpha-helical periodicity.

    The correct excel settings file template is necessary. Typically, the template is saved in the folder containing the
    scripts in the scanmut puthon module.

    Parameters
    ----------
    settings_file : excel file
        Path to excel file containing the scanmut settings.
            "settings" tab:
                List of datafiles to be processed, individual parameters associated with each file
                Each datafile should at LEAST contain a sample name, measured value, and optionally a y-error value.
                Datafiles can also contain columns with the original amino acid, position, and final amino acid. If
                this info is not available directly, you can get it from the dictionary on the settings excel file, or
                by automatic recognition of the mutation, based on the sample name.
            "dictionary" tab:
                 Columns forming a dictionary that links long sample names, e.g. insulin_precursor_Y108C, to separate
                 columns containing the relevant information ("original amino acid" = Y, "amino acid position" = 108,
                "final amino acid" = c.
            "explanations" tab:
                Explanation of elements in the excel file. Not used in the python script.

    show_barchart : boolean
        If True, the output figure will be shown as a popup window, or in IPython/Jupyter.
        For Ipython/Jupyter it is recommended to precede scanmut with the magic command %matplotlib inline.

    Saved Files and Figures
    -------
    All output files are saved in a subfolder based on the location of the input xlsx file:
    D:\Path\To\Your\Input\SettingsFile\scanmut_output\

    barchart : png, pdf
        Shows individual scanning mutagenesis data, clustered for each amino acid position, with bars coloured by
        the final amino acid.
    fit_sin_to_data_output : png, pdf, csv
        see tlabassays.sin_disrupt_fit.fit_sin_to_data for details regarding the fit-to-sine output files

    Usage
    -------
    import tlabtools
    settings_file = r"D:\Path\To\Your\Input\SettingsFile\scanmut_output\scanmut_settings.xlsx"
    tlabtools.run_scanmut(settings_file)
    """
    # open the settings as a pandas dataframe
    df_settings = pd.read_excel(settings_file, sheet_name = "settings")
    # fill any empty values (np.nan) with empty strings
    df_settings.fillna("", inplace=True)
    # open the dictionary showing the samplename and associated aa position as a pandas dataframe, dfd
    dfd = pd.read_excel(settings_file, sheetname = "dictionary", index_col="sample name")

    # normalise infile path to suit the operating system
    df_settings.loc[:,"infile"] = df_settings.loc[:,"infile"].apply(lambda x: os.path.normpath(x))

    for row in df_settings[df_settings.run_scanmut == True].index:
        sys.stdout.write("\n{}\n".format(df_settings.loc[row,"short experiment title"]))

        ###############################################################################################
        #                                                                                             #
        #         Read input file with scanmut data, define columns with data, etc                    #
        #                                                                                             #
        ###############################################################################################

        # define the path to the input file
        input_datafile = df_settings.loc[row,"infile"]
        # set up the output folder as the folder with the settings file, and a chosen subfolder
        outpath, outfile = eccpy.settings.setup_output_folder(settings_file, "scanmut_output")
        # obtain experiment title
        exp_title = df_settings.loc[row, "short experiment title"]
        outfile_base = outfile + "_" + exp_title
        # open the input datafile containing the scanning mutagenesis data
        if input_datafile[-4:] == ".csv":
            df = pd.read_csv(input_datafile, index_col = 0)
        elif input_datafile[-4:] == "xlsx":
            df = pd.read_excel(input_datafile, index_col=0)
        elif input_datafile[-4:] == ".xls":
            df = pd.read_excel(input_datafile, index_col=0)
        # raise an error if the dataframe is empty, due to a faulty excel file
        if df.empty:
            raise ValueError("The input file does not seem to contain data. Please check the filename, sheet_name,"
                             "and data contents.\nFile affected:\n{}".format(input_datafile))
        # set the index according to user input
        AAA_array = np.array(df_settings)

        if df_settings.loc[row, "index column"] != "":
            if df_settings.loc[row, "index column"] in df.columns:
                df.set_index(df_settings.loc[row, "index column"], drop=False, inplace=True)
        # define column with the y-axis data
        y_col = df_settings.loc[row,"y-data column"]
        # define column with the error bar data
        yerr_col = df_settings.loc[row, "y-error column"]
        # define final amino acid types (e.g. P in F255P, A256P) to be ignored
        list_ignored_final_aa = df_settings.loc[row,"ignored final amino acids"]
        # define the name of the wildtype
        wildtype_sample_name = df_settings.loc[row, "wildtype sample name"]

        ###############################################################################################
        #                                                                                             #
        #      Extract the amino acid positions, either from sample names or other columns            #
        #                                                                                             #
        ###############################################################################################

        # extract the amino acid position from the sample names
        if df_settings.loc[row,"amino-acid position identification method"] == "automatic":
            start_aa_pos = df_settings.loc[row,"start position"]
            end_aa_pos = df_settings.loc[row,"end position"]
            extract_aa_pos_from_sample_names(df, start_aa=start_aa_pos, end_aa=end_aa_pos, newcol_name="aa_pos")
        elif df_settings.loc[row,"amino-acid position identification method"] == "dictionary":
            # use the dictionary to extract the relevant amino acid positions from the sample names
            df = pd.concat([df, dfd.loc[df.index, :]], axis=1)
        elif "column_dictionary" in df_settings.loc[row,"amino-acid position identification method"]:
            column_dictionary = ast.literal_eval(df_settings.loc[row, "amino-acid position identification method"].split("=")[1])
            # if the dictionary contains the final columns as keys, reverse keys and values
            if "original amino acid" in column_dictionary.keys():
                column_dictionary = {v: k for k, v in column_dictionary.items()}
            # if all the columns in the dictionary are found in the dataframe, use dictionary to rename the columns
            if set(column_dictionary.keys()).issubset(set(df.columns)):
                df.rename(columns=column_dictionary, inplace=True)

        ###############################################################################################
        #                                                                                             #
        #      Setup wildtype, drop empty rows, create list of data and gap positions                 #
        #                                                                                             #
        ###############################################################################################

        # set the wildtype data as the first column to be shown
        if wildtype_sample_name != "":
            # set the amino acid position of the wildtype as 0, so that it sorts correctly
            df.loc[wildtype_sample_name, "amino acid position"] = df["amino acid position"].min() - 1
            df.loc[wildtype_sample_name, "original amino acid"] = "wt"
            df.loc[wildtype_sample_name, "final amino acid"] = "wt"

        # drop any of the columns that do not have a particular amino acid position (for example, positive control samples)
        df.dropna(subset=["amino acid position"], inplace=True)
        # drop any columns with mutations to amino acids that are excluded
        for final_aa in list_ignored_final_aa:
            df = df.loc[df['final amino acid'] != final_aa]

        # confirm that all of the amino acid positions are integers
        df["amino acid position"] = df["amino acid position"].astype(int)
        # sort the samples by the amino acid position in the dataframe
        df.sort_values(by=["amino acid position", "final amino acid"], inplace=True)
        # create a list of unique amino acid positions (for which there could be multiple mutations)
        list_unique_pos = df["amino acid position"].dropna().unique()
        # find the min and max of the amino acid positions
        min_aa_pos = list_unique_pos.min()
        max_aa_pos = list_unique_pos.max()
        # use min and max to create a contiguous list of positions, including any gaps
        list_contiguous_aa_pos = list(range(min_aa_pos,max_aa_pos+1))
        # create a list of gaps
        list_aa_pos_gaps = list(set(list_contiguous_aa_pos) - set(list_unique_pos))
        list_aa_pos_gaps.sort()
        if list_aa_pos_gaps != []:
            sys.stdout.write((", gaps (positions without data): " + str(list_aa_pos_gaps)))

        # set the index as the amino acid position
        df.set_index("amino acid position", inplace=True, drop=False)

        # double-check for errors, where the orig and final amino acid are the same
        errorseries = df["original amino acid"] == df["final amino acid"]
        if True in errorseries.value_counts():
            # collect the rows with the error
            error_rows = df.loc[errorseries].loc[:, ["original amino acid","final amino acid"]]
            # check if there is a wildtype row
            wtrow = error_rows[error_rows['original amino acid'] == "wt"].index.tolist()
            # drop the wildtype row from the error list
            error_rows.drop(wtrow, inplace=True)
            if not error_rows.empty:
                print("Warning, possible error found. Mutant listed as having wildtype sequence.\n""File:\n{}\n\nSample:"
                      "\n{}\n\n".format(input_datafile, error_rows))

        ###############################################################################################
        #                                                                                             #
        #  Move data to dfc_mult, which contains average values for all mutations, for fit to sine    #
        #                                                                                             #
        ###############################################################################################

        # create a new dataframe to store the average values for all mutations
        dfc_mult = pd.DataFrame()
        # create an empty list to hold the positions with no data (no mutants)
        for pos in list_contiguous_aa_pos:
            if pos in df.index:
                # select the row (or rows) with mutations at that position
                selected_data_at_aa_pos = df.loc[pos,:]
                # if selected_data_at_aa_pos is a series, there is only one mutation at that position.
                if isinstance(selected_data_at_aa_pos,pd.Series):
                    dfc_mult.loc[pos,"mean_mult"] = selected_data_at_aa_pos[y_col]
                    dfc_mult.loc[pos,"std_mult"] = 0
                # if selected_data_at_aa_pos is a dataframe, there are multiple mutations at that position. Take the mean.
                elif isinstance(selected_data_at_aa_pos,pd.DataFrame):
                    dfc_mult.loc[pos,"mean_mult"] = selected_data_at_aa_pos[y_col].mean()
                    dfc_mult.loc[pos,"std_mult"] = selected_data_at_aa_pos[y_col].std()
            else:
                # the position is not in the index, a gap is counted, the values will automatically be np.nan
                dfc_mult.loc[pos,"mean_mult"] = np.nan
                dfc_mult.loc[pos,"std_mult"] = 0

        if df_settings.loc[row, "run_fit_to_sin"] == True:
            #####################################################################################################
            #                                                                                                   #
            #                                        RUN FIT TO SINE                                            #
            #                                                                                                   #
            #####################################################################################################
            # define the x-axis for the sine calculation as the contiguous index integer
            x_sin = dfc_mult.index
            # the y-axis is the mean of all mutations at that position (note, NOT the mean of all replicates)
            y_sin = eccpytools.normalise_0_1(dfc_mult["mean_mult"])[0]
            title = "Scan. Mut. {}".format(exp_title[:25])
            output_tuple = fit_sin_to_data(x_sin, y_sin, title, outpath, outfile_base + "_sin")

            title_shuf = "Scan. Mut. {}, {}".format(exp_title[:25], "randomised")
            y_sin_shuffled = y_sin.copy()
            np.random.shuffle(y_sin_shuffled)
            output_tuple = fit_sin_to_data(x_sin, y_sin_shuffled, title_shuf, outpath, outfile_base  + "_sin" + "_randomised")

            print("Run sindisrupt is finished, files saved to {}".format(outpath))

        #####################################################################################################
        #                                                                                                   #
        #               CREATE BAR CHART WITH MUTATIONS TO MULTIPLE AMINO ACIDS                             #
        #                                                                                                   #
        #####################################################################################################

        # create a list of the final amino acids (ie., orig aa mutated to final aa)
        list_final_aa = df["final amino acid"].unique().tolist()
        # create empty list to hold the original amino acid
        list_orig_aa = []
        # create empty list to hold orig pos
        # list_orig_pos = []
        fontsize = 6
        plt.rcParams['font.size'] = fontsize
        # define width of bars in barchart, which will depend on number of bars per aa position
        max_n_aa_per_position = df["amino acid position"].value_counts().max()
        bar_width = 0.8 / max_n_aa_per_position
        # create figure and axes objects with matplotlib
        plt.close("all")
        fig_bar, ax_bar = plt.subplots()
        # create a list of colours, starting with black if the wildtype is included
        t20 = eccpy.tools.setup_t20_colour_list()
        # find the mutation number for that position, to place the bars next to each other in the barchart
        print(len(df.index.unique()))
        print(df.shape[0])
        for aa_pos in list_contiguous_aa_pos:
            if aa_pos in list_aa_pos_gaps:
                # NOTE some problem exists here with non-unique positions
                if len(df.index.unique()) == df.shape[0]:
                    df.loc[aa_pos,"mut_number_at_that_pos"] = 0
                    df.loc[aa_pos,"mut_offset_from_centre"] = 0
                    list_orig_aa.append("ND")
            # else:
            #     # add to a list of the actual positions with data
            #     list_orig_pos.append(aa_pos)
            else:
                # select data at that amino acid position, resulting in a dataframe or series (if only 1 mutant)
                df_series_aa_pos = df.loc[aa_pos,:]
                # if it's a series, there is only one mutant
                if isinstance(df_series_aa_pos, pd.Series):
                    df.loc[aa_pos,"mut_number_at_that_pos"] = 0
                    df.loc[aa_pos,"mut_offset_from_centre"] = 0
                    # select orig aa, add to list
                    list_orig_aa.append(df.loc[aa_pos,"original amino acid"])
                # if it's a dataframe, there are multiple mutants at that position
                elif isinstance(df_series_aa_pos, pd.DataFrame):
                    # count the mutants
                    n_mut = df_series_aa_pos.shape[0]
                    # insert the mutant number as a range of the number of mutants
                    df.loc[aa_pos,"mut_number_at_that_pos"] = range(n_mut)
                    offset = df.loc[aa_pos,"mut_number_at_that_pos"].mean()
                    df.loc[aa_pos,"mut_offset_from_centre"] = df.loc[aa_pos,"mut_number_at_that_pos"] - offset
                    list_orig_aa.append(df.loc[aa_pos,"original amino acid"].tolist()[0])

        # change the wildtype colour to grey, if it is in the dataset
        if list_orig_aa[0] == "wt":
            t20 = ["0.5"] + t20

        # define the general position of the bars on the x-axis, starting with 0
        df["barchart_index"] = df["amino acid position"] - min_aa_pos
        # define the exact position of each bar, based on the mutant number and calculated offset
        df["barchart_index_with_offset"] = df["barchart_index"] + df["mut_offset_from_centre"]*bar_width

        three_letter ={'V':'VAL', 'I':'ILE', 'L':'LEU', 'E':'GLU', 'Q':'GLN', \
        'D':'ASP', 'N':'ASN', 'H':'HIS', 'W':'TRP', 'F':'PHE', 'Y':'TYR',    \
        'R':'ARG', 'K':'LYS', 'S':'SER', 'T':'THR', 'M':'MET', 'A':'ALA',    \
        'G':'GLY', 'P':'PRO', 'C':'CYS'}

        # iterate through each final aa
        for n, aa in enumerate(list_final_aa):
            # set colour for those bars
            colour = t20[n]
            # select the data with that final aa
            df_final_aa = df.loc[df["final amino acid"] == aa]
            # obtain the position of the bars on the x-axis
            x_index = df_final_aa["barchart_index_with_offset"].tolist()
            # obtain the y-values and y-err
            y_values = df_final_aa[y_col].tolist()
            if yerr_col != "":
                y_err = df_final_aa[yerr_col].tolist()
            else:
                y_err = 0
            # define the legend-label
            if aa == "wt":
                legend_label = None
            else:
                legend_label = three_letter[aa]
            # plot the data
            bar_object = ax_bar.bar(x_index, y_values, color=colour, width=bar_width,
                                    yerr=y_err, error_kw=dict(ecolor='k', lw=1, capsize=2, capthick=1),
                                   label=legend_label, align="center")

        # obtain the number of datapoints on the x-axis
        n_dp = len(list_contiguous_aa_pos)
        # set the xticks
        xticks = np.array(range(n_dp))
        ax_bar.set_xticks(xticks)
        # create a series of the amino acid positions for the x-axis labels. Add the wt, if it is available.
        if list_orig_aa[0] == "wt":
            index = ["wt"] + list_contiguous_aa_pos[1:]
        else:
            index = list_contiguous_aa_pos
        # create a new dataframe to hold the xticklabels. use the list_contiguous_aa_pos as the index
        df_xticklabel = pd.DataFrame({"aa_pos": index,"list_orig_aa":list_orig_aa}, index=index).astype("U11")
        # for n, pos in enumerate(list_unique_pos):
        #     if pos in df_xticklabel.index:
        #         df_xticklabel.loc[pos, "aa"] = list_orig_aa[n]
        # df_xticklabel.fillna("", inplace=True)
        # df_xticklabel.index = df_xticklabel.index.fillna("")
        # df_xticklabel["xlabels"] = df_xticklabel.index  # + df_xticklabel["aa"]
        df_xticklabel["xlabels"] = df_xticklabel["aa_pos"].astype("U11") + "\n" + df_xticklabel["list_orig_aa"]
        ax_bar.set_xticklabels(df_xticklabel["xlabels"])
        # set the x-axis limits to be comfortable
        ax_bar.set_xlim(-bar_width*4, n_dp + bar_width*1)
        ax_bar.set_xlim(-1, n_dp + bar_width * 1)
        # set the y-axis limits as 5% above and below the max and min values, including error bars
        yerr_max = df[yerr_col].max() if yerr_col != "" else 0
        # extract the y-axis minimum or max from the settings file. Otherwise, pad ~10% around the min and max datapoints.
        if df_settings.loc[row,"ymax"] != "":
            ymax = df_settings.loc[row,"ymax"]
        else:
            ymax = (df[y_col].max() + yerr_max) * 1.1
        if df_settings.loc[row,"ymin"] != "":
            ymin = df_settings.loc[row,"ymin"]
        else:
            ymin = (df[y_col].min() - yerr_max) * 0.9
        ax_bar.set_ylim(ymin, ymax)
        # set the legend across the top (number of columns = number of final amino acids).
        ax_bar.legend(borderaxespad=2, ncol=len(list_final_aa))
        # set the y-axis title
        ax_bar.set_ylabel(df_settings.loc[row,"y-axis label"])
        # set title
        ax_bar.set_title(exp_title)
        fig_basename = os.path.join(outpath, outfile)
        fig_bar.savefig(fig_basename + "_" + exp_title + "_scanmut_barchart.png", format='png', dpi=150)
        if save_pdf == True:
            fig_bar.savefig(fig_basename + "_" + exp_title + "_scanmut_barchart.pdf", format='pdf')
        if show_barchart == True:
            plt.show()
        if df_settings.loc[row, "run_fit_to_sin"] != True:
            sys.stdout.write(", finished")

    print("Scanning mutagenesis analysis is finished.\nFiles saved to {}".format(outpath))
